(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: FishEye.cpp
//
// Desc: Example code showing how to do a fisheye lens effect with cubemapping.
//       The scene is rendering into a cubemap each frame, and then a
//       funky-shaped object is rendered using the cubemap and an environment
//       map.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit FishEyeUnit;

{$I DirectX.inc}

interface

uses
  Windows, SysUtils,
  Direct3D9, D3DX9, D3DApp, D3DEnumeration, D3DSettings, D3DFont, D3DFile,
  D3DUtil, DXUtil;

type
  //-----------------------------------------------------------------------------
  // Name: struct ENVMAPPEDVERTEX
  // Desc: D3D vertex type for environment-mapped objects
  //-----------------------------------------------------------------------------
  PEnvMappedVertex = ^TEnvMappedVertex;
  TEnvMappedVertex = packed record
    p: TD3DXVector3; // Position
    n: TD3DXVector3; // Normal
  end;

const
  TEnvMappedVertex_FVF = D3DFVF_XYZ or D3DFVF_NORMAL;

type
  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
    // Font for drawing text
    m_pFont: CD3DFont;

    m_pSkyBoxObject: CD3DMesh;
    m_pCubeMap: IDirect3DCubeTexture9;
    m_pRenderTargetOneSide: IDirect3DSurface9;

    m_pFishEyeLensVB: IDirect3DVertexBuffer9;
    m_pFishEyeLensIB: IDirect3DIndexBuffer9;
    m_dwNumFishEyeLensVertices: DWord;
    m_dwNumFishEyeLensFaces: DWord;

    function RenderScene: HResult;
    function RenderSceneIntoCubeMap: HResult;
    function GenerateFishEyeLens(dwNumRings: DWORD; dwNumSections: DWORD; fScale: Single): HResult;

  protected
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function Render: HResult; overload; override;
    function FrameMove: HResult; override;
    function FinalCleanup: HResult; override;

    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;
  public
    constructor Create;
  end;

implementation

{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create;

  m_strWindowTitle    := 'FishEye: Environment mapping';
  m_d3dEnumeration.AppUsesDepthBuffer := True;

  m_pFont                := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pSkyBoxObject        := CD3DMesh.Create;
  m_pCubeMap             := nil;
  m_pRenderTargetOneSide := nil;
  m_pFishEyeLensVB       := nil;
  m_pFishEyeLensIB       := nil;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HResult;
begin
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HResult;
var
  quat: TD3DXQuaternion;
  matTrackBall: TD3DXMatrixA16;
begin
  // When the window has focus, let the mouse adjust the scene
  if GetFocus <> 0 then
  begin
    quat:= D3DUtil_GetRotationFromCursor(m_hWnd);
    D3DXMatrixRotationQuaternion(matTrackBall, quat);
    m_pd3dDevice.SetTransform(D3DTS_WORLD, matTrackBall);

    // Render the scene into the surfaces of the cubemap
    if FAILED(RenderSceneIntoCubeMap) then
    begin
      Result:= E_FAIL;
      Exit;
    end;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RenderScene()
// Desc: Renders all visual elements in the scene. This is called by the main
//       Render() function, and also by the RenderIntoCubeMap() function.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RenderScene: HResult;
var
  matViewSave, matProjSave: TD3DXMatrixA16;
  matView,     matProj:     TD3DXMatrixA16;
begin
  // Render the skybox
  begin
    // Save current state
    m_pd3dDevice.GetTransform(D3DTS_VIEW,       matViewSave);
    m_pd3dDevice.GetTransform(D3DTS_PROJECTION, matProjSave);

    // Disable zbuffer, center view matrix, and set FOV to 90 degrees
    matView := matViewSave;
    matProj := matViewSave;
    matView._41:= 0; matView._42:= 0; matView._43:= 0;
    D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/2, 1.0, 0.5, 10000.0);
    m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);
    m_pd3dDevice.SetTransform(D3DTS_VIEW,       matView);
    m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,  iFalse);

    // Render the skybox
    m_pSkyBoxObject.Render(m_pd3dDevice);

    // Restore the render states
    m_pd3dDevice.SetTransform(D3DTS_VIEW,       matViewSave);
    m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProjSave);
    m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,  iTrue);
  end;

  // Render any other elements of the scene here. In this sample, only a
  // skybox is render, but a much more interesting scene could be rendered
  // instead.

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RenderSceneIntoCubeMap()
// Desc: Renders the scene to each of the 6 faces of the cube map
//-----------------------------------------------------------------------------
function CMyD3DApplication.RenderSceneIntoCubeMap: HResult;
var
  matViewSave, matProjSave: TD3DXMatrixA16;
  matView,     matProj:     TD3DXMatrixA16;
  matViewDir: TD3DXMatrixA16;

  pBackBuffer, pZBuffer: IDirect3DSurface9;
  i: TD3DCubemapFaces;
  pCubeMapFace: IDirect3DSurface9;
begin
  // Save transformation matrices of the device
  m_pd3dDevice.GetTransform(D3DTS_VIEW,       matViewSave);
  m_pd3dDevice.GetTransform(D3DTS_PROJECTION, matProjSave);

  // Set the projection matrix for a field of view of 90 degrees
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/2, 1.0, 0.5, 100.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  // Get the current view matrix, to concat it with the cubemap view vectors
  m_pd3dDevice.GetTransform(D3DTS_VIEW, matViewDir);
  matViewDir._41 := 0.0; matViewDir._42 := 0.0; matViewDir._43 := 0.0;

  // Store the current backbuffer and zbuffer
  m_pd3dDevice.GetRenderTarget(0, pBackBuffer);

  // Render to the six faces of the cube map
  for i:= D3DCUBEMAP_FACE_POSITIVE_X {0} to D3DCUBEMAP_FACE_NEGATIVE_Z {5} do
  begin
    // Set the view transform for this cubemap surface
    matView := D3DUtil_GetCubeMapViewMatrix(i);
    D3DXMatrixMultiply(matView, matViewDir, matView);
    m_pd3dDevice.SetTransform(D3DTS_VIEW, matView);

    // Set the rendertarget to our own RT surface
    m_pd3dDevice.SetRenderTarget(0, m_pRenderTargetOneSide);

    // Render the scene
    m_pd3dDevice.BeginScene;
    RenderScene;
    m_pd3dDevice.EndScene;

    // Copy the rendered image from our RT surface to the texture face
    m_pCubeMap.GetCubeMapSurface(i, 0, pCubeMapFace);
    m_pd3dDevice.StretchRect(m_pRenderTargetOneSide, nil, pCubeMapFace, nil, D3DTEXF_NONE);
    pCubeMapFace := nil;
  end;

  // Change the rendertarget back to the main backbuffer
  m_pd3dDevice.SetRenderTarget(0, pBackBuffer);
  pBackBuffer:= nil;
  pZBuffer:= nil;

  // Restore the original transformation matrices
  m_pd3dDevice.SetTransform(D3DTS_VIEW,       matViewSave);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProjSave);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HResult;
var
  matWorld: TD3DXMatrix;
begin
  // Begin the scene
  if FAILED(m_pd3dDevice.BeginScene) then
  begin
    Result:= S_OK;
    Exit;
  end;

  // Set the states we want: identity matrix, no z-buffer, and cubemap texture
  // coordinate generation
  D3DXMatrixIdentity(matWorld);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE, iFalse);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_CAMERASPACEREFLECTIONVECTOR);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT3);

  // Render the fisheye lens object with the environment-mapped body.
  m_pd3dDevice.SetTexture(0, m_pCubeMap);
  m_pd3dDevice.SetFVF(TEnvMappedVertex_FVF);
  m_pd3dDevice.SetStreamSource(0, m_pFishEyeLensVB, 0, SizeOf(TEnvMappedVertex));
  m_pd3dDevice.SetIndices(m_pFishEyeLensIB);
  m_pd3dDevice.DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0,
                                    0, m_dwNumFishEyeLensVertices,
                                    0, m_dwNumFishEyeLensFaces);

  // Restore the render states
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_PASSTHRU);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE);
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE, iTRUE);

  // Output statistics
  m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
  m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

  // End the scene.
  m_pd3dDevice.EndScene;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HResult;
begin
  // Initialize the font's internal textures
  m_pFont.InitDeviceObjects(m_pd3dDevice);

  // Load the file objects
  Result:= m_pSkyBoxObject.Create_(m_pd3dDevice, PChar('Lobby_skyboxInv.x'));
  if FAILED(Result) then
  begin
    Result:= D3DAPPERR_MEDIANOTFOUND;
    Exit;
  end;

  // Create the fisheye lens
  Result:= GenerateFishEyeLens(20, 20, 1.0);
  if FAILED(Result) then Exit;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restore device-memory objects and state after a device is created or
//       resized.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HResult;
var
  m_vEyePt:     TD3DXVector3;
  m_vLookatPt:  TD3DXVector3;
  m_vUpVec:     TD3DXVector3;

  matWorld, matView, matProj: TD3DXMatrixA16;
begin
  m_pFont.RestoreDeviceObjects;

  // Restore device objects for the skybox
  m_pSkyBoxObject.RestoreDeviceObjects(m_pd3dDevice);

  // Create the cubemap
  Result:= m_pd3dDevice.CreateCubeTexture(256, 1, D3DUSAGE_RENDERTARGET,
    m_d3dsdBackBuffer.Format, D3DPOOL_DEFAULT, m_pCubeMap, nil);
  if FAILED(Result) then Exit;

  // Create the render target for rendering cube faces
  Result:= m_pd3dDevice.CreateRenderTarget(256, 256, m_d3dsdBackBuffer.Format,
    m_d3dSettings.MultisampleType, m_d3dSettings.MultisampleQuality, False,
    m_pRenderTargetOneSide, nil);
  if FAILED(Result) then Exit;

  // Set default render states
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_ADDRESSU,  D3DTADDRESS_MIRROR);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_ADDRESSV,  D3DTADDRESS_MIRROR);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_NONE);

  // Set the transforms
  m_vEyePt    := D3DXVector3(0.0, 0.0,-5.0);
  m_vLookatPt := D3DXVector3(0.0, 0.0, 0.0);
  m_vUpVec    := D3DXVector3(0.0, 1.0, 0.0);

  D3DXMatrixIdentity(matWorld);
  D3DXMatrixLookAtLH(matView, m_vEyePt, m_vLookatPt, m_vUpVec);
  D3DXMatrixOrthoLH(matProj, 2.0, 2.0, 0.5, 100.0);
  m_pd3dDevice.SetTransform(D3DTS_WORLD,      matWorld);
  m_pd3dDevice.SetTransform(D3DTS_VIEW,       matView);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the device-dependent objects are about to be lost.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HResult;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pSkyBoxObject.InvalidateDeviceObjects;

  SAFE_RELEASE(m_pCubeMap);
  SAFE_RELEASE(m_pRenderTargetOneSide);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HResult;
begin
  m_pFont.DeleteDeviceObjects;
  m_pSkyBoxObject.Destroy_;

  m_pFishEyeLensVB:= nil;
  m_pFishEyeLensIB:= nil;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HResult;
begin
  FreeAndNil(m_pFont);
  FreeAndNil(m_pSkyBoxObject);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT;
begin
  Result:= E_FAIL;

  // GetTransform doesn't work on PUREDEVICE
  if ((dwBehavior and D3DCREATE_PUREDEVICE) <> 0) then Exit;

  // Check for cubemapping devices
  if (0 = (pCaps.TextureCaps and D3DPTEXTURECAPS_CUBEMAP)) then Exit;

  // Check that we can create a cube texture that we can render into
  if FAILED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal,
      pCaps.DeviceType, adapterFormat, D3DUSAGE_RENDERTARGET,
      D3DRTYPE_CUBETEXTURE, backBufferFormat))
  then Exit;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: GenerateFishEyeLens()
// Desc: Makes vertex and index data for a fish eye lens
//-----------------------------------------------------------------------------
function CMyD3DApplication.GenerateFishEyeLens(dwNumRings,
  dwNumSections: DWORD; fScale: Single): HResult;
var
  pVertices: PEnvMappedVertex;
  pIndices: PWord;
  dwNumTriangles: DWord;
  dwNumVertices:  DWord;
  i, j, m: DWord;
  r: Single;
  phi: Single;
  x, y, z: Single;
  nx, ny, nz: Single;
begin
  dwNumTriangles := (dwNumRings+1) * dwNumSections * 2;
  dwNumVertices  := (dwNumRings+1) * dwNumSections + 2;

  Result:= E_FAIL;

  // Generate space for the required triangles and vertices.
  if FAILED(m_pd3dDevice.CreateVertexBuffer(dwNumVertices*SizeOf(TEnvMappedVertex),
       D3DUSAGE_WRITEONLY, TEnvMappedVertex_FVF,
       D3DPOOL_MANAGED, m_pFishEyeLensVB, nil))
  then Exit;

  if FAILED(m_pd3dDevice.CreateIndexBuffer(dwNumTriangles*3*sizeof(WORD),
      D3DUSAGE_WRITEONLY, D3DFMT_INDEX16, D3DPOOL_MANAGED, m_pFishEyeLensIB, nil))
  then Exit;

  m_pFishEyeLensVB.Lock(0, 0, Pointer(pVertices), 0);
  m_pFishEyeLensIB.Lock(0, 0, Pointer(pIndices), 0);

  // Generate vertices at the end points.
  pVertices.p := D3DXVector3(0.0, 0.0, fScale);
  pVertices.n := D3DXVector3(0.0, 0.0, 1.0);
  Inc(pVertices);

  // Generate vertex points for rings
  r:= 0.0;
  for i:= 0 to dwNumRings do
  begin
    phi := 0.0;

    for j:= 0 to (dwNumSections - 1) do
    begin
      x :=  r * Sin(phi);
      y :=  r * Cos(phi);
      z := 0.5 - 0.5 * (x*x + y*y);

      nx:= -x;
      ny:= -y;
      nz:= 1.0;

      pVertices.p := D3DXVector3(x, y, z);
      pVertices.n := D3DXVector3(nx, ny, nz);
      Inc(pVertices);

      phi:= phi + (2*D3DX_PI / dwNumSections);
    end;

    r:= r + 1.5/dwNumRings;
  end;

  // Generate triangles for the centerpiece
  for i:= 0 to (2*dwNumSections - 1) do
  begin
    pIndices^:= (0);
    Inc(pIndices);
    pIndices^:= (i + 1);
    Inc(pIndices);
    pIndices^:= (1 + ((i + 1) mod dwNumSections));
    Inc(pIndices);
  end;

  // Generate triangles for the rings
  m := 1;  // 1st vertex begins at 1 to skip top point

  for i := 0 to (dwNumRings - 1) do
  begin
    for j := 0 to (dwNumSections - 1) do
    begin
      pIndices^:= (m + j);
      Inc(pIndices);
      pIndices^:= (m + dwNumSections + j);
      Inc(pIndices);
      pIndices^:= (m + dwNumSections + ((j + 1) mod dwNumSections));
      Inc(pIndices);

      pIndices^:= (m + j);
      Inc(pIndices);
      pIndices^:= (m + dwNumSections + ((j + 1) mod dwNumSections));
      Inc(pIndices);
      pIndices^:= (m + ((j + 1) mod dwNumSections));
      Inc(pIndices);
    end;
    m:= m + dwNumSections;
  end;

  m_pFishEyeLensVB.Unlock;
  m_pFishEyeLensIB.Unlock;
  m_dwNumFishEyeLensVertices := dwNumVertices;
  m_dwNumFishEyeLensFaces    := dwNumTriangles;

  Result:= S_OK;
end;

end.

