(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: SphereMap.cpp
//
// Desc: Example code showing how to use sphere-mapping in D3D, using generated
//       texture coordinates.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit SphereMapUnit;

interface

uses
  Windows, Messages, SysUtils,
  Direct3D9, D3DX9, D3DApp, D3DEnumeration, D3DSettings, D3DFile, D3DFont,
  D3DUtil, DXUtil;

//-----------------------------------------------------------------------------
// Name: g_szEffect
// Desc: String containing effect used to render shiny teapot.
//-----------------------------------------------------------------------------

const
  g_szEffect: PChar = // [] =

    'texture texSphereMap;'#10 +
    'matrix matWorld;'#10 +
    'matrix matViewProject;'#10 +
    'vector vecPosition;'#10 +

    'technique Sphere'#10 +
    '{'#10 +
        'pass P0'#10 +
        '{'#10 +

            // Vertex state
            'VertexShader ='#10 +
                'decl'#10 +
                '{'#10 +
                   // Decls no longer associated with vertex shaders in DX9
                '}'#10 +
                'asm'#10 +
                '{'#10 +
                    'vs_1_1'#10 +
                    'def c64, 0.25f, 0.5f, 1.0f, -1.0f'#10 +

                    'dcl_position v0'#10 +
                    'dcl_normal v1'#10 +

                    // r0: camera-space position
                    // r1: camera-space normal
                    // r2: camera-space vertex-eye vector
                    // r3: camera-space reflection vector
                    // r4: texture coordinates

                    // Transform position and normal into camera-space
                    'm4x4 r0, v0, c0'#10 +
                    'm3x3 r1.xyz, v1, c0'#10 +
                    'mov r1.w, c64.z'#10 +

                    // Compute normalized view vector
                    'add r2, c8, -r0'#10 +
                    'dp3 r3, r2, r2'#10 +
                    'rsq r3, r3.w'#10 +
                    'mul r2, r2, r3'#10 +

                    // Compute camera-space reflection vector
                    'dp3 r3, r1, r2'#10 +
                    'mul r1, r1, r3'#10 +
                    'add r1, r1, r1'#10 +
                    'add r3, r1, -r2'#10 +

                    // Compute sphere-map texture coords
                    'mad r4.w, -r3.z, c64.y, c64.y'#10 +
                    'rsq r4, r4.w'#10 +
                    'mul r4, r3, r4'#10 +
                    'mad r4, r4, c64.x, c64.y'#10 +

                    // Project position
                    'm4x4 oPos, r0, c4'#10 +
                    'mul oT0.xy, r4.xy, c64.zw'#10 +
                    'mov oT0.zw, c64.z'#10 +
                '};'#10 +

            'VertexShaderConstant4[0] = <matWorld>;'#10 +
            'VertexShaderConstant4[4] = <matViewProject>;'#10 +
            'VertexShaderConstant1[8] = <vecPosition>;'#10 +

            // Pixel state
            'Texture[0] = <texSphereMap>;'#10 +
            'AddressU[0] = Wrap;'#10 +
            'AddressV[0] = Wrap;'#10 +
            'MinFilter[0] = Linear;'#10 +
            'MagFilter[0] = Linear;'#10 +
            'ColorOp[0] = SelectArg1;'#10 +
            'ColorArg1[0] = Texture;'#10 +
        '}'#10 +
    '}';

var
  g_cchEffect: LongWord = 0; // = StrLen(g_szEffect) - 1;

//-----------------------------------------------------------------------------
// Name: struct ENVMAPPEDVERTEX
// Desc: D3D vertex type for environment-mapped objects
//-----------------------------------------------------------------------------
type
  PEnvMappedVertex = ^TEnvMappedVertex;
  TEnvMappedVertex = packed record
    p: TD3DXVector3; // Position
    n: TD3DXVector3; // Normal
  end;

const
  TEnvMappedVertex_FVF = D3DFVF_XYZ or D3DFVF_NORMAL;

type
  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_bCapture: Boolean;

    m_matProject        : TD3DXMatrix;
    m_matView           : TD3DXMatrix;
    m_matWorld          : TD3DXMatrix;
    m_matTrackBall      : TD3DXMatrix;

    m_pFont             : CD3DFont;
    m_pShinyTeapot      : CD3DMesh;
    m_pSkyBox           : CD3DMesh;

    m_pEffect           : ID3DXEffect;
    m_pSphereMap        : IDirect3DTexture9;
  protected
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function Render: HResult; overload; override;
    function FrameMove: HResult; override;
    function FinalCleanup: HResult; override;
  public
    constructor Create;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM ): LRESULT; override;
  end;

implementation

{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create;

  m_strWindowTitle    := 'SphereMap: Environment Mapping Technique';
  m_d3dEnumeration.AppUsesDepthBuffer := True;
  m_bCapture          := False;

  m_pFont             := nil;
  m_pShinyTeapot      := nil;
  m_pSkyBox           := nil;

  m_pEffect           := nil;
  m_pSphereMap        := nil;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HResult;
begin
  D3DXMatrixIdentity(m_matWorld);

  m_pFont        := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pShinyTeapot := CD3DMesh.Create;
  m_pSkyBox      := CD3DMesh.Create;

  if (m_pFont = nil) or (m_pShinyTeapot = nil) or (m_pSkyBox = nil) then
  begin
    Result:= E_OUTOFMEMORY;
    Exit;
  end;

  D3DXMatrixIdentity(m_matTrackBall);
  D3DXMatrixTranslation(m_matView, 0.0, 0.0, 3.0);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HResult;
var
  matCursor: TD3DXMatrix;
  qCursor: TD3DXQuaternion;
  matTrans: TD3DXMatrix;
  matRotation: TD3DXMatrix;
begin
  // When the window has focus, let the mouse adjust the camera view
  if m_bCapture then
  begin
    qCursor := D3DUtil_GetRotationFromCursor(m_hWnd);
    D3DXMatrixRotationQuaternion(matCursor, qCursor );
    D3DXMatrixMultiply(m_matView, m_matTrackBall, matCursor);

    D3DXMatrixTranslation(matTrans, 0.0, 0.0, 3.0);
    D3DXMatrixMultiply(m_matView, m_matView, matTrans);
  end else
  begin
    D3DXMatrixRotationY(matRotation, -m_fElapsedTime);
    D3DXMatrixMultiply(m_matWorld, m_matWorld, matRotation);
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HResult;
var
  matWorld, matView: TD3DXMatrix;
  matViewProject: TD3DXMatrix;
  matViewInv: TD3DXMatrix;
  vecPosition: TD3DXVector4;

  pVB: IDirect3DVertexBuffer9;
  pIB: IDirect3DIndexBuffer9;
  uPasses, iPass: LongWord;
begin
  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    // Render the Skybox
    begin
      D3DXMatrixScaling(matWorld, 10.0, 10.0, 10.0);

      matView:= m_matView;
      matView._41 := 0.0; matView._42 := 0.0; matView._43 := 0.0;

      m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);
      m_pd3dDevice.SetTransform(D3DTS_VIEW, matView);
      m_pd3dDevice.SetTransform(D3DTS_PROJECTION, m_matProject);

      m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
      m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);
      m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
      m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
      if ((m_d3dCaps.TextureAddressCaps and D3DPTADDRESSCAPS_MIRROR) = D3DPTADDRESSCAPS_MIRROR) then
      begin
        m_pd3dDevice.SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_MIRROR);
        m_pd3dDevice.SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_MIRROR);
      end;

      // Always pass Z-test, so we can avoid clearing color and depth buffers
      m_pd3dDevice.SetRenderState(D3DRS_ZFUNC, D3DCMP_ALWAYS);
      m_pSkyBox.Render(m_pd3dDevice);
      m_pd3dDevice.SetRenderState(D3DRS_ZFUNC, D3DCMP_LESSEQUAL);
    end;


    // Render the environment-mapped ShinyTeapot
    begin
      // Set transform state
      D3DXMatrixMultiply(matViewProject, m_matView, m_matProject);

      D3DXMatrixInverse(matViewInv, nil, m_matView);
      vecPosition:= D3DXVector4(matViewInv._41, matViewInv._42, matViewInv._43, 1.0);

      m_pEffect.SetMatrix('matWorld', m_matWorld);
      m_pEffect.SetMatrix('matViewProject', matViewProject);
      m_pEffect.SetVector('vecPosition', vecPosition);


      // Draw teapot
      m_pShinyTeapot.m_pLocalMesh.GetVertexBuffer(pVB);
      m_pShinyTeapot.m_pLocalMesh.GetIndexBuffer(pIB);

      m_pd3dDevice.SetFVF(m_pShinyTeapot.m_pLocalMesh.GetFVF);

      m_pd3dDevice.SetStreamSource(0, pVB, 0, SizeOf(TEnvMappedVertex));
      m_pd3dDevice.SetIndices(pIB);

      m_pEffect._Begin(@uPasses, 0);

      for iPass := 0 to uPasses - 1 do
      begin
        m_pEffect.Pass(iPass);

        m_pd3dDevice.DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 
          0, m_pShinyTeapot.m_pLocalMesh.GetNumVertices,
          0, m_pShinyTeapot.m_pLocalMesh.GetNumFaces);
      end;

      m_pEffect._End;

      SAFE_RELEASE(pVB);
      SAFE_RELEASE(pIB);
    end;


    // Output statistics
    m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);


    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HResult;
begin
  // Restore the device-dependent objects
  m_pFont.InitDeviceObjects(m_pd3dDevice);

  Result:= D3DAPPERR_MEDIANOTFOUND;

  // Load the file objects
  if FAILED(m_pShinyTeapot.Create_(m_pd3dDevice, PChar('Teapot.x'))) then Exit;
  if FAILED(m_pSkyBox.Create_(m_pd3dDevice, PChar('lobby_skybox.x'))) then Exit;
  if FAILED(D3DUtil_CreateTexture(m_pd3dDevice, 'Spheremap.bmp', m_pSphereMap)) then Exit;

  // Set mesh properties
  m_pShinyTeapot.SetFVF(m_pd3dDevice, TEnvMappedVertex_FVF);

  // Create Effect object
  if FAILED(D3DXCreateEffect(m_pd3dDevice, g_szEffect, g_cchEffect, nil, nil, 0, nil, m_pEffect, nil)) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restore device-memory objects and state after a device is created or
//       resized.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HResult;
var
  fAspect: Single;
begin
  // InitDeviceObjects for file objects (build textures and vertex buffers)
  m_pShinyTeapot.RestoreDeviceObjects(m_pd3dDevice);
  m_pSkyBox.RestoreDeviceObjects(m_pd3dDevice);
  m_pFont.RestoreDeviceObjects;
  m_pEffect.OnResetDevice;
  m_pEffect.SetTexture('texSphereMap', m_pSphereMap);

  // Set the transform matrices
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(m_matProject, D3DX_PI * 0.4, fAspect, 0.5, 100.0);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HResult;
begin
  m_pShinyTeapot.InvalidateDeviceObjects;
  m_pSkyBox.InvalidateDeviceObjects;
  m_pFont.InvalidateDeviceObjects;

  if Assigned(m_pEffect) then m_pEffect.OnLostDevice;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HResult;
begin
  m_pFont.DeleteDeviceObjects;
  m_pShinyTeapot.Destroy_;
  m_pSkyBox.Destroy_;

  SAFE_RELEASE(m_pSphereMap);
  SAFE_RELEASE(m_pEffect);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HResult;
begin
  SAFE_DELETE(m_pFont);
  SAFE_DELETE(m_pShinyTeapot);
  SAFE_DELETE(m_pSkyBox);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
  adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT;
begin
  if (dwBehavior and D3DCREATE_SOFTWARE_VERTEXPROCESSING = 0) and
      not (pCaps.VertexShaderVersion >= D3DVS_VERSION(1, 0)) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT;
var
  matCursor: TD3DXMatrix;
  qCursor: TD3DXQuaternion;
begin
  // Capture mouse when clicked
  if (WM_LBUTTONDOWN = uMsg) then
  begin
    qCursor := D3DUtil_GetRotationFromCursor(m_hWnd);
    D3DXMatrixRotationQuaternion(matCursor, qCursor);
    D3DXMatrixTranspose(matCursor, matCursor);
    D3DXMatrixMultiply(m_matTrackBall, m_matTrackBall, matCursor);

    SetCapture(m_hWnd);
    m_bCapture := True;
    Result:= 0;
    Exit;
  end;

  if (WM_LBUTTONUP = uMsg) then
  begin
    qCursor := D3DUtil_GetRotationFromCursor(m_hWnd);
    D3DXMatrixRotationQuaternion(matCursor, qCursor);
    D3DXMatrixMultiply(m_matTrackBall, m_matTrackBall, matCursor);

    ReleaseCapture;
    m_bCapture := False;
    Result:= 0;
    Exit;
  end;

  // Pass remaining messages to default handler
  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

begin
  g_cchEffect:= StrLen(g_szEffect);
end.


